/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
#ifdef WITH_STARPU
#include <starpu.h>
#include "common_pastix.h"
#include "starpu_kernels.h"
#include "sparse_gemm_blocktab.h"

#include <inttypes.h>

#ifdef STARPU_USE_CUDA
#if ((!defined FORCE_DOUBLE)  || (!(defined __CUDA_ARCH__) || __CUDA_ARCH__ >= 130))
#if !(defined FORCE_DOUBLE && defined TYPE_COMPLEX)
#ifndef FORCE_NO_CUDA
#define STARPU_USE_CUDA_GEMM_FUNC
#endif
#endif
#endif
#endif

#define DimTrans                API_CALL(DimTrans)
#define kernel_trsm             API_CALL(kernel_trsm)
#define CORE_gemdm              API_CALL(CORE_gemdm)

#if !(defined STARPU_USE_CUDA_GEMM_FUNC)
#define CUDA_SPARSE_GEMM_BLOCKTAB(TRANSA, TRANSB,       \
                                  dimi, dimj, dima,     \
                                  alpha,  A,  stride_A, \
                                  B,  stride_B,         \
                                  beta,  C, stride_C,   \
                                  blocknbr, blocktab,   \
                                  fblocknbr, fblocktab) \
  do {                                                  \
  } while(0)
#else
#define CUDA_SPARSE_GEMM_BLOCKTAB(TRANSA, TRANSB,       \
                                  dimi, dimj, dima,     \
                                  alpha,  A,  stride_A, \
                                  B,  stride_B,         \
                                  beta,  C, stride_C,   \
                                  blocknbr, blocktab,   \
                                  fblocknbr, fblocktab) \
  do {                                                                  \
    magmablas_sparse_gemm_blocktab_kernel_N_T_64_16_4_16_4(C, A, B,    \
                                                           (int)dimi,   \
                                                           (int)dimj,   \
                                                           (int)dima,   \
                                                           (int)stride_A, \
                                                           (int)stride_B, \
                                                           (int)stride_C, \
                                                           (float)alpha, \
                                                            (float)beta, \
                                                           blocknbr,    \
                                                           blocktab,    \
                                                           fblocknbr,   \
                                                           fblocktab);  \
  } while(0)
#endif


#define SOPALIN_SPARSE_GEMM(TRANSA, TRANSB,                     \
                            dimi, dimj, dima,                   \
                            alpha,  A,  stride_A,               \
                            B,  stride_B,                       \
                            beta,  C, stride_C,                 \
                            nblocs, blocs_idx, facing_bloc_idx, \
                            wtmp, wtmpsize)                     \
  do {                                                          \
    sparse_gemm(TRANSA, TRANSB,                                 \
                dimi, dimj, dima,                               \
                alpha,  A,  stride_A,                           \
                B,  stride_B,                                   \
                beta,  C, stride_C,                             \
                nblocs, blocs_idx, facing_bloc_idx,             \
                wtmp, wtmpsize);                                \
  } while(0)

#define SOPALIN_SPARSE_GEMM_BLOCKTAB(TRANSA, TRANSB,        \
                                     dimi, dimj, dima,      \
                                     alpha,  A,  stride_A,  \
                                     B,  stride_B,          \
                                     beta,  C, stride_C,    \
                                     blocknbr, blocktab,    \
                                     fblocknbr, fblocktab,  \
                                     wtmp, wtmpsize)        \
  do {                                                      \
    assert(0);                                              \
  } while(0)


#ifdef CHOL_SOPALIN
#ifdef SOPALIN_LU
/* General LU decomposition */
/*
  Function: getrfsp1d_starpu_cpu

  Diagonal block factorization and column block update for LU decomposition.

  Parameters:
    buffers - Data handlers :
      0 - Pivot counter
      1 - L column block
      2 - U column block
    _args   - Codelet arguments:
     sopalin_data - global PaStiX internal data.
     cblknum      - Current column block index.
*/
void getrfsp1d_starpu_cpu(void * buffers[], void * _args)
{
  starpu_trf_data_t * args         = (starpu_trf_data_t*)_args;
  Sopalin_Data_t    * sopalin_data = args->sopalin_data;
  SolverMatrix      * datacode     = sopalin_data->datacode;
  INT               * nbpivots     = (INT*)STARPU_VARIABLE_GET_PTR(buffers[0]);
  FLOAT             * lDiag        = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT             * uDiag        = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[2]);
  INT                 stride       = STARPU_MATRIX_GET_LD(buffers[1]);
  INT                 cblknum      = args->cblknum;
  INT                 fblknum      = SYMB_BLOKNUM(cblknum);
  INT                 lblknum      = SYMB_BLOKNUM(cblknum+1);
  FLOAT             * lExtraDiag   = NULL;
  FLOAT             * uExtraDiag   = NULL;
  INT                 dima         = SYMB_LCOLNUM(cblknum) -
    SYMB_FCOLNUM(cblknum) + 1;
  INT                 dimb         = stride - dima;

  /* check if diagonal column block */
  assert( SYMB_FCOLNUM(cblknum) == SYMB_FROWNUM(fblknum) );

  /* Add U diagonal updates into L */
  SOPALIN_GEAM("T", "N", dima, dima, 1.0,
               uDiag, stride,
               lDiag, stride);

  /* Factorize diagonal block (two terms version with workspace) */
  PASTIX_getrf_block(lDiag, dima, stride,
                     nbpivots,
                     sopalin_data->critere);
  /* Transpose L_diag in U_diag Matrix */
  DimTrans(lDiag,stride, dima,uDiag);

  /* if there is an extra-diagonal bloc in column block */
  if ( fblknum+1 < lblknum )
    {
      lExtraDiag = lDiag + SOLV_COEFIND(fblknum+1);
      uExtraDiag = uDiag + SOLV_COEFIND(fblknum+1);
      kernel_trsm(dimb, dima,
                  lDiag,       uDiag,      stride,
                  lExtraDiag,  uExtraDiag, stride);
    }
}

/*
  Function: getrfsp1d_gemm_starpu_common

  General LU update of left block column facing current block.

  Common function for CPU and GPU.

  Update block by block.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - U column block.
      3 - U facing column block.
      4 - Working memory area.
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
   arch     - indicate if the codelet is runned on CPU or CUDA node.
*/
static inline void
getrfsp1d_gemm_starpu_common(void * buffers[], void * _args, int arch)
{
  starpu_gemm_data_t * args         = (starpu_gemm_data_t*)_args;
  Sopalin_Data_t     * sopalin_data = args->sopalin_data;
  FLOAT              * L            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[0]);
  FLOAT              * Cl           = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT              * U            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[2]);
  FLOAT              * Cu           = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[3]);
  FLOAT              * work         = (FLOAT*)STARPU_VECTOR_GET_PTR(buffers[4]);
  INT                  stride       = STARPU_MATRIX_GET_LD(buffers[0]);
  INT                  stridefc     = STARPU_MATRIX_GET_LD(buffers[1]);
  INT                  cblknum      = args->cblknum;
  INT                  bloknum      = args->bloknum;
  INT                  fcblknum     = args->fcblknum;
  FLOAT              * wtmp;
  FLOAT              * C;
  FLOAT              * Aik;
  FLOAT              * Akj;
  FLOAT              * Aij;
  SolverMatrix       *datacode = sopalin_data->datacode;
  INT fblknum, lblknum, frownum;
  INT indblok;
  INT b, j;
  INT dimi, dimj, dima, dimb;

  fblknum = SYMB_BLOKNUM(cblknum);
  lblknum = SYMB_BLOKNUM(cblknum + 1);

  indblok = SOLV_COEFIND(bloknum);

  dimi = stride - indblok;
  dimj = SYMB_LROWNUM(bloknum) - SYMB_FROWNUM(bloknum) + 1;
  dima = STARPU_MATRIX_GET_NY(buffers[0]);

  /* Matrix A = Aik */
  Aik = L + indblok;
  Akj = U + indblok;
  wtmp = work;

  switch(arch) {
  case ARCH_CPU:
    /* fprintf(stdout, */
    /*         "L  %016"   PRIXPTR " U  %016"  PRIXPTR */
    /*         " Cl  %016" PRIXPTR " Cu  %016" PRIXPTR "\n", L, U, Cl, Cu); */

    SOPALIN_GEMM( "N", "T",
                  dimi, dimj, dima,
                  1.,  Aik,  stride,
                  Akj,  stride,
                  0.,  wtmp, dimi);
    break;
  default:
    errorPrint("Unknown Architecture");
    assert(0);
    break;
  }

  /*
   * Add contribution to facing cblk
   */
  b = SYMB_BLOKNUM( fcblknum );
  C = Cl + (SYMB_FROWNUM(bloknum) - SYMB_FCOLNUM(fcblknum)) * stridefc;
  /* for all following blocks in block column */
  for (j=bloknum; j<lblknum; j++) {
    frownum = SYMB_FROWNUM(j);

    /* Find facing bloknum */
    while (!BLOCK_ISFACING(j,b))
      {
        b++;
        assert( b < SYMB_BLOKNUM( fcblknum+1 ) );
      }

    Aij = C + SOLV_COEFIND(b) + frownum - SYMB_FROWNUM(b);
    dimb = SYMB_LROWNUM(j) - frownum + 1;

    switch(arch) {
    case ARCH_CPU:
      SOPALIN_GEAM("N", "N", dimb, dimj, -1.0,
                   wtmp, dimi,
                   Aij,  stridefc );
      break;
    default:
      errorPrint("Unknown Architecture");
      assert(0);
      break;
    }


    /* Displacement to next block */
    wtmp += dimb;
  }

  /*
   * Compute update on U
   */

  Aik = U + indblok;
  Akj = L + indblok;
  wtmp = work;
  switch(arch) {
  case ARCH_CPU:
    SOPALIN_GEMM( "N", "T",
                  dimi, dimj, dima,
                  1.,  Aik,  stride,
                  Akj,  stride,
                  0.,  wtmp, dimi  );
    break;
  default:
    errorPrint("Unknown Architecture");
    assert(0);
    break;
  }

  wtmp += SYMB_LROWNUM(bloknum) - SYMB_FROWNUM(bloknum) + 1;

  /*
   * Add contribution to facing cblk
   */
  b = SYMB_BLOKNUM( fcblknum );
  C = Cl + (SYMB_FROWNUM(bloknum) - SYMB_FCOLNUM(fcblknum));

  /* for all following blocks in block column */
  for (j=bloknum+1; j<lblknum; j++) {
    frownum = SYMB_FROWNUM(j);

    /* Find facing bloknum */
    /* WARNING: may not work for NAPA */
    if (
#ifdef NAPA_SOPALIN /* ILU(k) */
        !(((SYMB_FROWNUM(j)>=SYMB_FROWNUM(b)) &&
           (SYMB_LROWNUM(j)<=SYMB_LROWNUM(b))) ||
          ((SYMB_FROWNUM(j)<=SYMB_FROWNUM(b)) &&
           (SYMB_LROWNUM(j)>=SYMB_LROWNUM(b))) ||
          ((SYMB_FROWNUM(j)<=SYMB_FROWNUM(b)) &&
           (SYMB_LROWNUM(j)>=SYMB_FROWNUM(b))) ||
          ((SYMB_FROWNUM(j)<=SYMB_LROWNUM(b)) &&
           (SYMB_LROWNUM(j)>=SYMB_LROWNUM(b))))
#else
        !((SYMB_FROWNUM(j)>=SYMB_FROWNUM(b)) &&
          (SYMB_LROWNUM(j)<=SYMB_LROWNUM(b)))
#endif
        )
      break;


    Aij = C + (frownum - SYMB_FROWNUM(b))*stridefc;
    dimb = SYMB_LROWNUM(j) - frownum + 1;

    switch(arch) {
    case ARCH_CPU:
      SOPALIN_GEAM( "T", "N", dimj, dimb, -1.0,
                    wtmp, dimi,
                    Aij,  stridefc );
      break;
    default:
      errorPrint("Unknown Architecture");
      assert(0);
      break;
    }

    /* Displacement to next block */
    wtmp += dimb;
  }


  C = Cu + (SYMB_FROWNUM(bloknum) - SYMB_FCOLNUM(fcblknum)) * stridefc;

  /* Keep updating on U */
  for (; j<lblknum; j++) {
    frownum = SYMB_FROWNUM(j);

    /* Find facing bloknum */
    while (!BLOCK_ISFACING(j,b))
      {
        b++;
        assert( b < SYMB_BLOKNUM( fcblknum+1 ) );
      }

    dimb = SYMB_LROWNUM(j) - frownum + 1;
    Aij = C + SOLV_COEFIND(b) + frownum - SYMB_FROWNUM(b);
    switch(arch) {
    case ARCH_CPU:
      SOPALIN_GEAM("N", "N", dimb, dimj, -1.0,
                   wtmp, dimi,
                   Aij,  stridefc );
      break;
    default:
      errorPrint("Unknown Architecture");
      assert(0);
      break;
    }


    /* Displacement to next block */
    wtmp += dimb;
  }
}

/*
  Function: getrfsp1d_gemm_starpu_cpu

  General LU update of left block column facing current block.

  Update block by block.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - U column block.
      3 - U facing column block.
      4 - Working memory area.
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
*/
void getrfsp1d_gemm_starpu_cpu(void * buffers[], void * _args)
{
  getrfsp1d_gemm_starpu_common(buffers, _args, ARCH_CPU);
}

#ifdef STARPU_USE_CUDA
/*
  Function: getrfsp1d_gemm_starpu_cuda

  General LU update of left block column facing current block.

  Update block by block.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - U column block.
      3 - U facing column block.
      4 - Working memory area.
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
*/
void getrfsp1d_gemm_starpu_cuda(void * buffers[], void * _args)
{
  getrfsp1d_gemm_starpu_common(buffers, _args, ARCH_CUDA);
}
#endif


/*
  Function: getrfsp1d_gemm_starpu_common

  Sparse LU update of left block column facing current block.

  Common function for CPU and GPU.

  Update all the facing column block at once.

  TODO: Implement the CPU version

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - U column block.
      3 - U facing column block.
      4 - Working memory area.
      5 - blocktab (depending on -DSTARPU_BLOCKTAB_SELFCOPY).
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
      nblocs       - Number of blocks in current column block.
      d_blocktab   - pointers on blocktabs for CUDA nodes
                     (-DSTARPU_BLOCKTAB_SELFCOPY).
   arch     - indicate if the codelet is runned on CPU or CUDA node.
*/
static inline
void getrfsp1d_sparse_gemm_starpu_common(void * buffers[], void * _args,
                                         int arch)
{
  starpu_gemm_data_t * args         = (starpu_gemm_data_t*)_args;
  Sopalin_Data_t     * sopalin_data = args->sopalin_data;
  FLOAT              * L            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[0]);
  FLOAT              * Cl           = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT              * U            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[2]);
  FLOAT              * Cu           = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[3]);
  FLOAT              * work         = (FLOAT*)STARPU_VECTOR_GET_PTR(buffers[4]);
  size_t               worksize     = STARPU_VECTOR_GET_NX(buffers[4]);
#ifdef STARPU_BLOCKTAB_SELFCOPY
  int                  device;
  int                * all_blocktab;
#else
  int                * all_blocktab = (int*)STARPU_VECTOR_GET_PTR(buffers[5]);
#endif
  INT                  stride       = STARPU_MATRIX_GET_LD(buffers[0]);
  INT                  stridefc     = STARPU_MATRIX_GET_LD(buffers[1]);
  INT                  cblknum      = args->cblknum;
  INT                  bloknum      = args->bloknum;
  INT                  fcblknum     = args->fcblknum;
  SolverMatrix       * datacode     = sopalin_data->datacode;
  INT                  indblok      = SOLV_COEFIND(bloknum);
  INT                  dimi         = stride - indblok;
  INT                  dimj         = BLOK_ROWNBR(bloknum);
  INT                  dima         = CBLK_COLNBR(cblknum);
  FLOAT                alpha        = -1.;
  FLOAT                beta         = 1.;
  FLOAT              * Aik;
  FLOAT              * Akj;
  FLOAT              * Aij;
  int                  blocknbr  = SYMB_BLOKNUM(cblknum+1) - bloknum;
  int                  fblocknbr =
    SYMB_BLOKNUM(fcblknum+1) - SYMB_BLOKNUM(fcblknum);
#ifdef STARPU_BLOCKTAB_SELFCOPY
  int                * blocktab;
  int                * fblocktab;
  if (arch == ARCH_CUDA) {
    cudaGetDevice(&device);
    all_blocktab=args->d_blocktab[device];
    blocktab   = &(all_blocktab[2*bloknum]);
    fblocktab  = &(all_blocktab[2*SYMB_BLOKNUM(fcblknum)]);
  }
#else
  int                * blocktab  = &(all_blocktab[2*bloknum]);
  int                * fblocktab = &(all_blocktab[2*SYMB_BLOKNUM(fcblknum)]);
#endif

  Aik = L + indblok;
  Akj = U + indblok;
  Aij = Cl;

  switch(arch) {
  case ARCH_CPU:
    Aij = Aij + (SYMB_FROWNUM(bloknum) -
                 SYMB_FCOLNUM(fcblknum))*SOLV_STRIDE(fcblknum);

    SOPALIN_SPARSE_GEMM_BLOCKTAB( "N", "T",
                                  dimi, dimj, dima,
                                  alpha,  Aik,  stride,
                                  Akj,  stride,
                                  beta,  Aij, stridefc,
                                  blocknbr, blocktab, fblocknbr, fblocktab,
                                  work, worksize);
    break;
#ifdef STARPU_USE_CUDA
  case ARCH_CUDA:
    Aij = Aij + (SYMB_FROWNUM(bloknum) -
                 SYMB_FCOLNUM(fcblknum))*SOLV_STRIDE(fcblknum);
    CUDA_SPARSE_GEMM_BLOCKTAB( "n", "t",
                               dimi, dimj, dima,
                               alpha,  Aik,  stride,
                               Akj,  stride,
                               beta,  Aij, stridefc,
                               blocknbr, blocktab, fblocknbr, fblocktab);
    break;
#endif
  default:
    errorPrint("Unknown Architecture");
    assert(0);
    break;
  }

  /*
   * Compute update on U
   */
  if ( blocknbr > 1 )
    {
      dimi = dimi - (SOLV_COEFIND(bloknum+1) - SOLV_COEFIND(bloknum));
      Aik = U + SOLV_COEFIND(bloknum+1);
      Akj = L + indblok;
      Aij = Cu;
      switch(arch) {
      case ARCH_CPU:
        Aij = Aij + (SYMB_FROWNUM(bloknum) -
                     SYMB_FCOLNUM(fcblknum))*SOLV_STRIDE(fcblknum);

        SOPALIN_SPARSE_GEMM_BLOCKTAB( "N", "T",
                                      dimi, dimj, dima,
                                      alpha,  Aik,  stride,
                                      Akj,  stride,
                                      beta,  Aij, stridefc,
                                      blocknbr-1, &(blocktab[2]),
                                      fblocknbr, fblocktab,
                                      work, worksize);
  break;
#ifdef STARPU_USE_CUDA
      case ARCH_CUDA:
        Aij = Aij + (SYMB_FROWNUM(bloknum) -
                     SYMB_FCOLNUM(fcblknum))*SOLV_STRIDE(fcblknum);
        CUDA_SPARSE_GEMM_BLOCKTAB( "n", "t",
                                   dimi, dimj, dima,
                                   alpha,  Aik,  stride,
                                   Akj,  stride,
                                   beta,  Aij, stridefc,
                                   blocknbr-1, &(blocktab[2]),
                                   fblocknbr, fblocktab);
  break;
#endif
      default:
        errorPrint("Unknown Architecture");
        assert(0);
        break;
      }
    }
}


/*
  Function: getrfsp1d_gemm_starpu_cpu

  Sparse LU update of left block column facing current block.

  Update all the facing column block at once.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - U column block.
      3 - U facing column block.
      4 - Working memory area.
      5 - blocktab (depending on -DSTARPU_BLOCKTAB_SELFCOPY).
    _args   - codelet arguments 
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
      nblocs       - Number of blocks in current column block.
      d_blocktab   - pointers on blocktabs for CUDA nodes
                     (-DSTARPU_BLOCKTAB_SELFCOPY).
*/
void getrfsp1d_sparse_gemm_starpu_cpu(void * buffers[], void * _args)
{
  getrfsp1d_gemm_starpu_common(buffers, _args, ARCH_CPU);
}

/*
  Function: getrfsp1d_gemm_starpu_cuda

  Sparse LU update of left block column facing current block.

  Update all the facing column block at once.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - U column block.
      3 - U facing column block.
      4 - Working memory area.
      5 - blocktab (depending on -DSTARPU_BLOCKTAB_SELFCOPY).
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
      nblocs       - Number of blocks in current column block.
      d_blocktab   - pointers on blocktabs for CUDA nodes
                     (-DSTARPU_BLOCKTAB_SELFCOPY).
*/
void getrfsp1d_sparse_gemm_starpu_cuda(void * buffers[], void * _args)
{
  getrfsp1d_sparse_gemm_starpu_common(buffers, _args, ARCH_CUDA);
}

#else /* not SOPALIN_LU */
/* LLT decomposition */
/*
  Function: potrfsp1d_starpu_cpu

  Diagonal block factorization and column block update for LLt decomposition.

  Parameters:
    buffers - Data handlers :
      0 - Pivot counter
      1 - L column block
    _args   - Codelet arguments:
     sopalin_data - global PaStiX internal data.
     cblknum      - Current column block index.
*/
void
potrfsp1d_starpu_cpu(void * buffers[], void * _args)
{
  starpu_trf_data_t * args         = (starpu_trf_data_t*)_args;
  Sopalin_Data_t    * sopalin_data = args->sopalin_data;
  SolverMatrix      * datacode     = sopalin_data->datacode;
  INT               * nbpivots     = (INT*)STARPU_VARIABLE_GET_PTR(buffers[0]);
  FLOAT             * Diag         = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  INT                 stride       = STARPU_MATRIX_GET_LD(buffers[1]);
  INT                 cblknum      = args->cblknum;
  INT                 fblknum      = SYMB_BLOKNUM(cblknum);
  INT                 lblknum      = SYMB_BLOKNUM(cblknum+1);
  INT                 dima         = CBLK_COLNBR(cblknum);
  INT                 dimb         = stride - dima;
  FLOAT             * ExtraDiag    = NULL;


  /* check if diagonal column block */
  assert( SYMB_FCOLNUM(cblknum) == SYMB_FROWNUM(fblknum) );

  /* Factorize diagonal block (two terms version with workspace) */
  PASTIX_potrf_block(Diag, dima, stride,
                      nbpivots,
                      sopalin_data->critere);

  /* if there is an extra-diagonal bloc in column block */
  if ( fblknum+1 < lblknum )
    {
      ExtraDiag    = Diag + SOLV_COEFIND(fblknum+1);
      kernel_trsm(dimb, dima,
                  Diag,      stride,
                  ExtraDiag, stride);
    }
}

/*
  Function: potrfsp1d_gemm_starpu_cpu

  General LLt update of left block column facing current block.

  Update block by block.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - Working memory area.
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
*/
void potrfsp1d_gemm_starpu_cpu(void * buffers[], void * _args)
{
  starpu_gemm_data_t * args         = (starpu_gemm_data_t*)_args;
  Sopalin_Data_t     * sopalin_data = args->sopalin_data;
  SolverMatrix       * datacode     = sopalin_data->datacode;
  FLOAT              * L            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[0]);
  FLOAT              * C            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT              * work         = (FLOAT*)STARPU_VECTOR_GET_PTR(buffers[2]);
  INT                  stride       = STARPU_MATRIX_GET_LD(buffers[0]);
  INT                  stridefc     = STARPU_MATRIX_GET_LD(buffers[1]);
  INT                  cblknum      = args->cblknum;
  INT                  bloknum      = args->bloknum;
  INT                  fcblknum     = args->fcblknum;
  INT                  lblknum      = SYMB_BLOKNUM(cblknum+1);
  INT                  indblok      = SOLV_COEFIND(bloknum);
  INT                  dimi         = stride - indblok;
  INT                  dimj         = BLOK_ROWNBR(bloknum);
  INT                  dima         = CBLK_COLNBR(cblknum);

  FLOAT *Aik, *Aij;
  INT frownum;
  INT b, j;
  INT dimb;


  /* Matrix A = Aik */
  Aik = L + indblok;

  /* Compute the contribution */
  SOPALIN_GEMM( "N", "C",
                dimi, dimj, dima,
                1.,  Aik,  stride,
                Aik,  stride,
                0.,  work, dimi);

  /*
   * Add contribution to facing cblk
   */
  b = SYMB_BLOKNUM( fcblknum );
  C = C + (SYMB_FROWNUM(bloknum) - SYMB_FCOLNUM(fcblknum)) * stridefc;
  /* for all following blocks in block column */
  for (j=bloknum; j<lblknum; j++) {
    frownum = SYMB_FROWNUM(j);

    /* Find facing bloknum */
    while(!BLOCK_ISFACING(j,b))
      {
        b++;
        assert( b < SYMB_BLOKNUM( fcblknum+1 ) );
      }

    Aij = C + SOLV_COEFIND(b) + frownum - SYMB_FROWNUM(b);
    dimb = SYMB_LROWNUM(j) - frownum + 1;

    MAT_zaxpy( dimb, dimj, -1.0,
               work, dimi,
               Aij,  stridefc );

    /* Displacement to next block */
    work += dimb;
  }
}

/*
  Function: potrfsp1d_gemm_starpu_common

  Sparse LLt update of left block column facing current block.

  Common function for CPU and GPU.

  Update all the facing column block at once.

  TODO: Implement the CPU version

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - Working memory area.
      3 - blocktab (depending on -DSTARPU_BLOCKTAB_SELFCOPY).
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
      nblocs       - Number of blocks in current column block.
      d_blocktab   - pointers on blocktabs for CUDA nodes
                     (-DSTARPU_BLOCKTAB_SELFCOPY).
    arch    - indicate if the codelet is runned on CPU or CUDA node.
*/
static inline
void potrfsp1d_sparse_gemm_starpu_common(void * buffers[], void * _args,
                                         int arch)
{
  starpu_gemm_data_t * args         = (starpu_gemm_data_t*)_args;
  Sopalin_Data_t     * sopalin_data = args->sopalin_data;
  SolverMatrix       * datacode     = sopalin_data->datacode;
  FLOAT              * L            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[0]);
  FLOAT              * C            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT              * work         = (FLOAT*)STARPU_VECTOR_GET_PTR(buffers[2]);
  size_t               worksize     = STARPU_VECTOR_GET_NX(buffers[2]);
  INT                  stride       = STARPU_MATRIX_GET_LD(buffers[0]);
  INT                  stridefc     = STARPU_MATRIX_GET_LD(buffers[1]);
  int                  device;
#ifdef STARPU_BLOCKTAB_SELFCOPY
  int                * all_blocktab;
#else
  int                * all_blocktab = (int*)STARPU_VECTOR_GET_PTR(buffers[3]);
#endif
  INT                  cblknum      = args->cblknum;
  INT                  bloknum      = args->bloknum;
  INT                  fcblknum     = args->fcblknum;
  INT                  indblok      = SOLV_COEFIND(bloknum);
  INT                  dimi         = stride - indblok;
  INT                  dimj         = BLOK_ROWNBR(bloknum);
  INT                  dima         = CBLK_COLNBR(cblknum);
  FLOAT                alpha = -1.;
  FLOAT                beta = 1.;
  FLOAT              * Aik;
  FLOAT              * Aij;
  int                  blocknbr     = SYMB_BLOKNUM(cblknum+1) - bloknum;
  int                  fblocknbr    = CBLK_BLOKNBR(fcblknum);
#ifdef STARPU_BLOCKTAB_SELFCOPY
  int                * blocktab;
  int                * fblocktab;
  cudaGetDevice(&device);
  all_blocktab=args->d_blocktab[device];
  blocktab   = &(all_blocktab[2*bloknum]);
  fblocktab  = &(all_blocktab[2*SYMB_BLOKNUM(fcblknum)]);
#else
  int                * blocktab     = &(all_blocktab[2*bloknum]);
  int                * fblocktab    = &(all_blocktab[2*SYMB_BLOKNUM(fcblknum)]);
#endif

  Aik = L + indblok;
  Aij = C;
  /* Compute the contribution */
  switch(arch) {
  case ARCH_CPU:
    Aij = Aij + (SYMB_FROWNUM(bloknum) -
                 SYMB_FCOLNUM(fcblknum))*SOLV_STRIDE(fcblknum);

    SOPALIN_SPARSE_GEMM_BLOCKTAB( "N", "C",
                                  dimi, dimj, dima,
                                  alpha,  Aik,  stride,
                                  Aik,  stride,
                                  beta,  Aij, stridefc,
                                  blocknbr, blocktab, fblocknbr, fblocktab,
                                  work, worksize);
    break;
#ifdef STARPU_USE_CUDA
  case ARCH_CUDA:
    Aij = Aij + (SYMB_FROWNUM(bloknum) -
                 SYMB_FCOLNUM(fcblknum))*SOLV_STRIDE(fcblknum);

    CUDA_SPARSE_GEMM_BLOCKTAB( "n", "c",
                               dimi, dimj, dima,
                               alpha,  Aik,  stride,
                               Aik,  stride,
                               beta,  Aij, stridefc,
                               blocknbr, blocktab, fblocknbr, fblocktab);
    break;
#endif
  default:
    errorPrint("Unknown Architecture");
    assert(0);
    break;
  }

}

/*
  Function: potrfsp1d_gemm_starpu_cpu

  Sparse LLt update of left block column facing current block.

  Update all the facing column block at once.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - Working memory area.
      3 - blocktab (depending on -DSTARPU_BLOCKTAB_SELFCOPY).
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
      nblocs       - Number of blocks in current column block.
      d_blocktab   - pointers on blocktabs for CUDA nodes
                     (-DSTARPU_BLOCKTAB_SELFCOPY).
*/
void potrfsp1d_sparse_gemm_starpu_cpu(void * buffers[], void * _args)
{
  potrfsp1d_gemm_starpu_cpu(buffers, _args);
}

/*
  Function: potrfsp1d_gemm_starpu_cuda

  Sparse LLt update of left block column facing current block.

  Update all the facing column block at once.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - Working memory area.
      3 - blocktab (depending on -DSTARPU_BLOCKTAB_SELFCOPY).
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
      nblocs       - Number of blocks in current column block.
      d_blocktab   - pointers on blocktabs for CUDA nodes
                     (-DSTARPU_BLOCKTAB_SELFCOPY).
*/
void potrfsp1d_sparse_gemm_starpu_cuda(void * buffers[], void * _args)
{
  potrfsp1d_sparse_gemm_starpu_common(buffers, _args, ARCH_CUDA);
}

#endif /* not SOPALIN_LU */
#else  /* not CHOL_SOPALIN */
/* LDLT Decomposition */
/*
  Function: hetrfsp1d_starpu_cpu

  Diagonal block factorization and column block update for LDLt decomposition.

  Parameters:
    buffers - Data handlers :
      0 - Pivot counter
      1 - L column block
      2 - Will receive L*DIAG(BLOCK_DIAG(L))
    _args   - Codelet arguments:
     sopalin_data - global PaStiX internal data.
     cblknum      - Current column block index.
*/
void hetrfsp1d_starpu_cpu(void * buffers[], void * _args)
{
  starpu_trf_data_t * args         = (starpu_trf_data_t*)_args;
  Sopalin_Data_t    * sopalin_data = args->sopalin_data;
  SolverMatrix      * datacode     = sopalin_data->datacode;
  INT               * nbpivots     = (INT*)STARPU_VARIABLE_GET_PTR(buffers[0]);
  FLOAT             * Diag         = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT             * tmp4         = (FLOAT*)STARPU_VECTOR_GET_PTR(buffers[2]);
  INT                 stride       = STARPU_MATRIX_GET_LD(buffers[1]);
  INT                 cblknum      = args->cblknum;
  INT                 fblknum      = SYMB_BLOKNUM(cblknum);
  INT                 lblknum      = SYMB_BLOKNUM(cblknum+1);
  FLOAT             * ExtraDiag    = NULL;
  INT                 dima         = CBLK_COLNBR(cblknum);
  INT                 dimb         = stride - dima;

  /* check if diagonal column block */
  assert( SYMB_FCOLNUM(cblknum) == SYMB_FROWNUM(fblknum) );

  /* Initialisation des pointeurs de blocs */

  /* Factorize diagonal block (two terms version with workspace) */
#ifdef HERMITIAN
  PASTIX_hetrf_block(Diag, dima, stride,
                     nbpivots,
                     sopalin_data->critere,
                     tmp4);
#else
  PASTIX_sytrf_block(Diag, dima, stride,
                     nbpivots,
                     sopalin_data->critere,
                     tmp4);
#endif
  /* if there is an extra-diagonal bloc in column block */
  if ( fblknum+1 < lblknum )
    {
      ExtraDiag = Diag + SOLV_COEFIND(fblknum+1);
      /* WARNING : Second "tmp4" should be a buffer to use for send. */
      kernel_trsm(dimb, dima,
                  Diag,      stride,
                  ExtraDiag, tmp4, stride);
    }
}

/*
  Function: hetrfsp1d_gemm_starpu_cpu

  General LDLt update of left block column facing current block.

  Common function for CPU and GPU.

  Update block by block.

  Parameters:
    buffers - Data handlers :
      0 - L column block.
      1 - L facing column block.
      2 - Working memory area.
    _args   - codelet arguments :
      sopalin_data - global PaStiX internal data.
      cblknum      - Current column block index.
      bloknum      - Current block index.
      fcblknum     - Facing column block index.
*/
void hetrfsp1d_gemm_starpu_cpu(void * buffers[], void * _args)
{
  starpu_gemm_data_t * args         = (starpu_gemm_data_t*)_args;
  Sopalin_Data_t     * sopalin_data = args->sopalin_data;
  SolverMatrix       * datacode     = sopalin_data->datacode;
  FLOAT              * L            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[0]);
  FLOAT              * C            = (FLOAT*)STARPU_MATRIX_GET_PTR(buffers[1]);
  FLOAT              * work1        = (FLOAT*)STARPU_VECTOR_GET_PTR(buffers[2]);
  FLOAT              * work2        = work1 + SOLV_COEFMAX;
  INT                  cblknum      = args->cblknum;
  INT                  bloknum      = args->bloknum;
  INT                  fcblknum     = args->fcblknum;
  FLOAT *Aik, *Aij;
  INT fblknum, lblknum, frownum;
  INT stride, stridefc, indblok;
  INT b, j;
  INT dimi, dimj, dima, dimb;
  INT ldw = SOLV_COEFMAX;
  fblknum = SYMB_BLOKNUM(cblknum);
  lblknum = SYMB_BLOKNUM(cblknum + 1);

  indblok = SOLV_COEFIND(bloknum);
  stride  = SOLV_STRIDE(cblknum);

  dimi = stride - indblok;
  dimj = SYMB_LROWNUM(bloknum) - SYMB_FROWNUM(bloknum) + 1;
  dima = SYMB_LCOLNUM(cblknum) - SYMB_FCOLNUM(cblknum) + 1;

  /* Matrix A = Aik */
  Aik = L + indblok;

  /* Compute the contribution */
  CORE_gemdm( PastixNoTrans, PastixConjTrans,
              dimi, dimj, dima,
              1.,  Aik,   stride,
              Aik,   stride,
              0.,  work1, dimi,
              L,     stride+1,
              work2, ldw );

  /*
   * Add contribution to facing cblk
   */
  b = SYMB_BLOKNUM( fcblknum );
  stridefc = SOLV_STRIDE(fcblknum);
  C = C + (SYMB_FROWNUM(bloknum) - SYMB_FCOLNUM(fcblknum)) * stridefc;

  /* for all following blocks in block column */
  for (j=bloknum; j<lblknum; j++) {
    frownum = SYMB_FROWNUM(j);

    /* Find facing bloknum */
    while (!BLOCK_ISFACING(j,b))
      {
        b++;
        assert( b < SYMB_BLOKNUM( fcblknum+1 ) );
      }


    Aij = C + SOLV_COEFIND(b) + frownum - SYMB_FROWNUM(b);
    dimb = SYMB_LROWNUM(j) - frownum + 1;

    MAT_zaxpy( dimb, dimj, -1.0,
               work1, dimi,
               Aij,   stridefc );

    /* Displacement to next block */
    work1 += dimb;
  }
}
#endif /* not CHOL_SOPALIN */

#else  /* not WITH_STARPU */
/* ISO C forbids an empty source file */
#include "not_empty.h"
NOT_EMPTY( starpu_kernels )
#endif /* not WITH_STARPU */
