/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
#ifdef WITH_STARPU
#  include <cuda.h>
#  include <starpu.h>
#  include <starpu_profiling.h>
#  ifdef FORCE_NOMPI
#    include "nompi.h"
#  else
#    include <mpi.h>
#  endif
#  include "common_pastix.h"
#  include "sopalin_define.h"
#  include "sopalin_acces.h"
#  include "symbol.h"
#  include "ftgt.h"
#  include "csc.h"
#  include "updown.h"
#  include "queue.h"
#  include "bulles.h"
#  include "solver.h"
#  include "sopalin_thread.h"
#  include "sopalin_time.h"
#  include "sopalin3d.h"
#  include "starpu_kernels.h"
#  include "../../perf/src/perf.h"
#  include "starpu_submit_tasks.h"

#  ifdef STARPU_USE_CUDA
#    if ((!defined FORCE_DOUBLE) || __CUDA_ARCH__ >= 130)
#      if !(defined FORCE_DOUBLE && defined TYPE_COMPLEX)
#        ifndef FORCE_NO_CUDA
#          define STARPU_USE_CUDA_GEMM_FUNC
#        endif
#      endif
#    endif
#  endif

#  define CUDA_CALL(x) do                                           \
    {                                                             \
      if (cudaSuccess != x)                                       \
        {                                                         \
          errorPrint("Problem getting number of CUDA devices\n"); \
          assert(0);                                              \
        }                                                         \
    } while(0)

void  dump_all(SolverMatrix *, int);

#  define USE_TASK_DEP

#  if (STARPU_MAJOR_VERSION < 1)
#    error "PaStiX requires STARPU >= 1"
#  endif /* STARPU_MAJOR_VERSION */

#  define SPARSE_GEMM
#  define ARCH_CPU  0
#  define ARCH_CUDA 1

static struct starpu_perfmodel init_pivot_model = {
  .type = STARPU_HISTORY_BASED,
  .symbol = "INIT_PIVOT"
};
static struct starpu_perfmodel redux_pivot_model = {
  .type = STARPU_HISTORY_BASED,
  .symbol = "REDUX_PIVOT"
};

static size_t trf_size(struct starpu_task *task,
                       enum starpu_perf_archtype arch,
                       unsigned nimpl)
{
  starpu_trf_data_t * args         = (starpu_trf_data_t*)task->cl_arg;
  Sopalin_Data_t    * sopalin_data = args->sopalin_data;
  SolverMatrix      * datacode     = sopalin_data->datacode;
  INT                 stride       = STARPU_MATRIX_GET_LD(task->handles[1]);
  size_t              dima         = CBLK_COLNBR(args->cblknum);
  return OPS_PPF(dima) + OPS_TRSM(dima, stride);

}

static size_t gemm_size(struct starpu_task *task,
                        enum starpu_perf_archtype arch,
                        unsigned nimpl)
{
  starpu_gemm_data_t         * args         = (starpu_gemm_data_t*)task->cl_arg;
  Sopalin_Data_t             * sopalin_data = args->sopalin_data;
  INT                          stride       = STARPU_MATRIX_GET_LD(task->handles[0]);
  INT                          bloknum      = args->bloknum;
  SolverMatrix               * datacode     = sopalin_data->datacode;
  INT                          indblok      = SOLV_COEFIND(bloknum);
  INT                          dimi         = stride - indblok;
  INT                          dimj         = BLOK_ROWNBR(bloknum);
  INT                          dima         = STARPU_MATRIX_GET_NY(task->handles[0]);
  size_t size = 0;

  size += OPS_GEMM(dimi,dimj,dima);
  return size;
}

static struct starpu_perfmodel GEMM_model = {
  .type = STARPU_HISTORY_BASED,
  .per_arch =
  {
    [STARPU_CPU_DEFAULT][0]  = { .size_base = gemm_size },
    [STARPU_CUDA_DEFAULT][0] = { .size_base = gemm_size }
  },
  .symbol = "GEMM"
};

#  define dump_all                               API_CALL(dump_all)
#  define starpu_partition_data                  API_CALL(starpu_partition_data)


#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
/* LU */
static struct starpu_perfmodel GETRF_TRSM_model = {
  .type = STARPU_HISTORY_BASED,
  .symbol = "GETRF_TRSM",
  .per_arch =
  {
    [STARPU_CPU_DEFAULT][0]  = { .size_base = trf_size },
    [STARPU_CUDA_DEFAULT][0] = { .size_base = trf_size }
  }
};

#      define getrfsp1d_cl             PASTIX_PREFIX_F(getrfsp1d_cl)
#      define getrfsp1d_gemm_cl        PASTIX_PREFIX_F(getrfsp1d_gemm_cl)
#      define getrfsp1d_sparse_gemm_cl PASTIX_PREFIX_F(getrfsp1d_sparse_gemm_cl)

struct starpu_codelet getrfsp1d_cl =
  {
    .where = STARPU_CPU,
    .cpu_func = getrfsp1d_starpu_cpu,
    .model = &GETRF_TRSM_model,
    .nbuffers = 3,
    .modes = {
      STARPU_REDUX,
      STARPU_RW,
      STARPU_RW}
  };

struct starpu_codelet getrfsp1d_gemm_cl =
  {
    .where = STARPU_CPU
#      ifdef STARPU_USE_CUDA_GEMM_FUNC
    |STARPU_CUDA
#      endif
    ,
    .cpu_func = getrfsp1d_gemm_starpu_cpu,
#      ifdef STARPU_USE_CUDA_GEMM_FUNC
    .cuda_func = getrfsp1d_gemm_starpu_cuda,
#      endif
    .model = &GEMM_model,
    .nbuffers = 5,
    .modes = {STARPU_R,
              STARPU_RW,
              STARPU_R,
              STARPU_RW,
              STARPU_SCRATCH}
  };

struct starpu_codelet getrfsp1d_sparse_gemm_cl =
  {
    .where = STARPU_CPU
#      ifdef STARPU_USE_CUDA_GEMM_FUNC
    |STARPU_CUDA
#      endif
    ,
    .cpu_func = getrfsp1d_sparse_gemm_starpu_cpu,
#      ifdef STARPU_USE_CUDA_GEMM_FUNC
    .cuda_func = getrfsp1d_sparse_gemm_starpu_cuda,
#      endif
    .model = &GEMM_model,
#      if (defined STARPU_BLOCKTAB_SELFCOPY)
    .nbuffers = 5,
#      else
    .nbuffers = 6,
#      endif
    .modes = {STARPU_R,
              STARPU_RW,
              STARPU_R,
              STARPU_RW,
              STARPU_SCRATCH,
#      if !(defined STARPU_BLOCKTAB_SELFCOPY)
              STARPU_R
#      endif
    }
  };
#      define cl_trf       getrfsp1d_cl
#      ifdef SPARSE_GEMM
#        define cl_gemm      getrfsp1d_sparse_gemm_cl
#      else
#        define cl_gemm      getrfsp1d_gemm_cl
#      endif
#    else /* SOPALIN_LU */
/* LLT */
static struct starpu_perfmodel POTRF_TRSM_model = {
  .type = STARPU_HISTORY_BASED,
  .symbol = "POTRF_TRSM",
  .per_arch =
  {
    [STARPU_CPU_DEFAULT][0]  = { .size_base = trf_size },
    [STARPU_CUDA_DEFAULT][0] = { .size_base = trf_size }
  }

};
#      define potrfsp1d_cl             PASTIX_PREFIX_F(potrfsp1d_cl)
#      define potrfsp1d_gemm_cl        PASTIX_PREFIX_F(potrfsp1d_gemm_cl)
#      define potrfsp1d_sparse_gemm_cl PASTIX_PREFIX_F(potrfsp1d_sparse_gemm_cl)
struct starpu_codelet potrfsp1d_cl =
  {
    .where = STARPU_CPU,
    .cpu_func = potrfsp1d_starpu_cpu,
    .model = &POTRF_TRSM_model,
    .nbuffers = 2,
    .modes = {
      STARPU_REDUX,
      STARPU_RW}
  };
struct starpu_codelet potrfsp1d_gemm_cl =
  {
    .where = STARPU_CPU,
    .cpu_func = potrfsp1d_gemm_starpu_cpu,
    .model = &GEMM_model,
    .nbuffers = 3,
    .modes = {STARPU_R,
              STARPU_RW,
              STARPU_SCRATCH}
  };
struct starpu_codelet potrfsp1d_sparse_gemm_cl =
  {
    .where = STARPU_CPU
#      ifdef STARPU_USE_CUDA_GEMM_FUNC
    | STARPU_CUDA
#      endif
    ,
    .cpu_func = potrfsp1d_sparse_gemm_starpu_cpu,
#      ifdef STARPU_USE_CUDA_GEMM_FUNC
    .cuda_func = potrfsp1d_sparse_gemm_starpu_cuda,
#      endif
    .model = &GEMM_model,
#      if (defined STARPU_BLOCKTAB_SELFCOPY)
    .nbuffers = 3,
#      else
    .nbuffers = 4,
#      endif

    .modes = {STARPU_R,
              STARPU_RW,
              STARPU_SCRATCH,
#      if !(defined STARPU_BLOCKTAB_SELFCOPY)
              STARPU_R
#      endif
    }
  };
#      define cl_trf       potrfsp1d_cl
#      ifdef SPARSE_GEMM
#        define cl_gemm      potrfsp1d_sparse_gemm_cl
#      else
#        define cl_gemm      potrfsp1d_gemm_cl
#      endif
#    endif /* SOPALIN_LU */
#  else  /* CHOL_SOPALIN */
/* LDLT */
static struct starpu_perfmodel HETRF_TRSM_model = {
  .type = STARPU_HISTORY_BASED,
  .symbol = "HETRF_TRSM",
  .per_arch =
  {
    [STARPU_CPU_DEFAULT][0]  = { .size_base = trf_size },
    [STARPU_CUDA_DEFAULT][0] = { .size_base = trf_size }
  }

};
#ifdef HERMITIAN
#    define hetrfsp1d_cl             PASTIX_PREFIX_F(hetrfsp1d_cl)
#    define hetrfsp1d_gemm_cl        PASTIX_PREFIX_F(hetrfsp1d_gemm_cl)
#    define hetrfsp1d_sparse_gemm_cl PASTIX_PREFIX_F(hetrfsp1d_sparse_gemm_cl)
#else
#    define hetrfsp1d_cl             PASTIX_PREFIX_F(sytrfsp1d_cl)
#    define hetrfsp1d_gemm_cl        PASTIX_PREFIX_F(sytrfsp1d_gemm_cl)
#    define hetrfsp1d_sparse_gemm_cl PASTIX_PREFIX_F(sytrfsp1d_sparse_gemm_cl)
#endif
struct starpu_codelet hetrfsp1d_cl =
  {
    .where = STARPU_CPU,
    .cpu_func = hetrfsp1d_starpu_cpu,
    .model = &HETRF_TRSM_model,
    .nbuffers = 3,
    .modes = {
      STARPU_REDUX,
      STARPU_RW,
      STARPU_SCRATCH}
  };

struct starpu_codelet hetrfsp1d_gemm_cl =
  {
    .where = STARPU_CPU,
    .cpu_func = hetrfsp1d_gemm_starpu_cpu,
    .model = &GEMM_model,
    .nbuffers = 3,
    .modes = {STARPU_R,
              STARPU_RW,
              STARPU_SCRATCH}
  };

struct starpu_codelet hetrfsp1d_sparse_gemm_cl =
  {
    .where = STARPU_CPU,
    .cpu_func = hetrfsp1d_gemm_starpu_cpu,
    .model = &GEMM_model,
#    if (defined STARPU_BLOCKTAB_SELFCOPY)
    .nbuffers = 3,
#    else
    .nbuffers = 4,
#    endif
    .modes = { STARPU_R,
	       STARPU_RW,
	       STARPU_SCRATCH,
#    if !(defined STARPU_BLOCKTAB_SELFCOPY)
	       STARPU_R
#    endif
    }
  };
#    define cl_trf       hetrfsp1d_cl
#    ifdef SPARSE_GEMM
#      define cl_gemm      hetrfsp1d_sparse_gemm_cl
#    else
#      define cl_gemm      hetrfsp1d_gemm_cl
#    endif
#  endif /* CHOL_SOPALIN */


/*
 *  Codelets to implement reduction
 */

/*
  Function: init_cpu_func

  Initialize the pivot counter on CPU.

  Parameters:
    descr  - The pivot data descriptor.
    cl_arg - The arguments of the codelet (void).
 */
static void init_cpu_func(void *descr[], void *cl_arg)
{
  INT *val = (INT *)STARPU_VARIABLE_GET_PTR(descr[0]);
  *val = 0;
}

#  ifdef STARPU_HAVE_CURAND
/*
  Function: init_cuda_func

  Initialize the pivot counter on CUDA nodes.

  Parameters:
    descr  - The pivot data descriptor.
    cl_arg - The arguments of the codelet (void).
 */
static void init_cuda_func(void *descr[], void *cl_arg)
{
  INT *val = (INT *)STARPU_VARIABLE_GET_PTR(descr[0]);
  cudaMemset(val, 0, sizeof(INT));
  cudaThreadSynchronize();
}
#  endif

/*
  Pivot initialization codelet.
*/
static struct starpu_codelet init_codelet = {
  .where =
#  ifdef STARPU_HAVE_CURAND
  STARPU_CUDA|
#  endif
  STARPU_CPU,
  .cpu_func = init_cpu_func,
#  ifdef STARPU_HAVE_CURAND
  .cuda_func = init_cuda_func,
#  endif
  .nbuffers = 1,
  .model    = &init_pivot_model
};

#  ifdef STARPU_HAVE_CURAND
/*
  Function: redux_cuda_func

  Dummy implementation of the addition of two unsigned longs in CUDA

  Parameters:
    descr  - The pivot data descriptor.
    cl_arg - The arguments of the codelet (void).
*/
static void redux_cuda_func(void *descr[], void *cl_arg)
{
  INT *d_a = (INT *)STARPU_VARIABLE_GET_PTR(descr[0]);
  INT *d_b = (INT *)STARPU_VARIABLE_GET_PTR(descr[1]);

  INT h_a, h_b;

  cudaMemcpy(&h_a, d_a, sizeof(h_a), cudaMemcpyDeviceToHost);
  cudaMemcpy(&h_b, d_b, sizeof(h_b), cudaMemcpyDeviceToHost);

  h_a += h_b;

  cudaMemcpy(d_a, &h_a, sizeof(h_a), cudaMemcpyHostToDevice);
};
#  endif
/*
  Function: redux_cpu_func

  Dummy implementation of the addition of two unsigned longs on CPUs.

  Parameters:
    descr  - The pivot data descriptor.
    cl_arg - The arguments of the codelet (void).
*/
static void redux_cpu_func(void *descr[], void *cl_arg)
{
  INT *a = (INT *)STARPU_VARIABLE_GET_PTR(descr[0]);
  INT *b = (INT *)STARPU_VARIABLE_GET_PTR(descr[1]);

  *a = *a + *b;
};


/*
  Pivot reduction codelet.
*/
static struct starpu_codelet redux_codelet = {
  .where =
#  ifdef STARPU_HAVE_CURAND
  STARPU_CUDA|
#  endif
  STARPU_CPU,
  .cpu_func = redux_cpu_func,
#  ifdef STARPU_HAVE_CURAND
  .cuda_func = redux_cuda_func,
#  endif
  .nbuffers = 2,
  .model    = &redux_pivot_model
};

/*
  Function: starpu_data_partition

  Initialize column blocks handlers.

  Parameters:
    sopalin_data - PaStiX global data structure.
    L_handle     - Handles for L column blocks.
    U_handle     - Handles for U column blocks.
*/
static void starpu_partition_data(Sopalin_Data_t * sopalin_data,
                                  starpu_data_handle_t * L_handle,
                                  starpu_data_handle_t * U_handle)
{
  SolverMatrix       * datacode         = sopalin_data->datacode;
  INT itertask;


  for (itertask=0;itertask<SOLV_TASKNBR;itertask++)
    {
      INT itercblk = TASK_CBLKNUM(itertask);
      INT nblocs   = CBLK_BLOKNBR(itercblk);
      INT coefnbr  = SOLV_STRIDE(itercblk) * CBLK_COLNBR(itercblk);
      INT i, blocnum;

      starpu_matrix_data_register(&(L_handle[itercblk]), 0,
                                  (uintptr_t)SOLV_COEFTAB(itercblk),
                                  (uint32_t)SOLV_STRIDE(itercblk),
                                  (uint32_t)SOLV_STRIDE(itercblk),
                                  CBLK_COLNBR(itercblk),
                                  sizeof(FLOAT));
#  if defined USE_TASK_DEP
      starpu_data_set_sequential_consistency_flag(L_handle[itercblk], 0);
#  endif

#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
      starpu_matrix_data_register(&(U_handle[itercblk]), 0,
                                  (uintptr_t)SOLV_UCOEFTAB(itercblk),
                                  (uint32_t)SOLV_STRIDE(itercblk),
                                  (uint32_t)SOLV_STRIDE(itercblk),
                                  CBLK_COLNBR(itercblk),
                                  sizeof(FLOAT));
#      if (defined USE_TASK_DEP)
      starpu_data_set_sequential_consistency_flag(U_handle[itercblk], 0);
#      endif
#    endif
#  endif
    }

}

/*
  Function: data_unregister_callback

  Unregister a data handler.

  Parameters:
    arg - pointer to the data handler.
*/
static void
data_unregister_callback(void* arg)
{
  starpu_data_handle_t * handle = (starpu_data_handle_t*)arg;
  starpu_data_unregister(*handle);
}


/*
  Function: starpu_submit_tasks

  Submit tasks to perform the decomposition of the matrix.

  Parameters:
    sopalin_data - PaStiX global data structure.

  Returns:
    NO_ERR
 */
int
starpu_submit_tasks(Sopalin_Data_t  * sopalin_data) {
  SolverMatrix         * datacode         = sopalin_data->datacode;
  Thread_Data_t        * thread_data;
  starpu_data_handle_t   pivot_cnt_handle;
  INT nbpivots;
  starpu_trf_data_t    * trf_args;
  starpu_gemm_data_t   * gemm_args;
  starpu_data_handle_t * L_handle;
#  ifdef USE_TASK_DEP
  struct starpu_task  ** starpu_tasktab;
#  endif
#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
  starpu_data_handle_t * U_handle;
#    endif
#  endif
  starpu_data_handle_t   WORK_handle;
  INT itertask, count;
  int * blocktab;
#  ifdef STARPU_BLOCKTAB_SELFCOPY
  int ** d_blocktab;
#  else
  starpu_data_handle_t   blocktab_handle;
#  endif

#  if (defined USE_TASK_DEP)
  INT deps_tab_size = 1024;
  struct starpu_task ** deps;
  MALLOC_INTERN(deps, deps_tab_size, struct starpu_task *);
#  endif

#ifdef STARPU_PROFILING
  starpu_profiling_status_set(STARPU_PROFILING_ENABLE);
#endif

  {
    /* build blocktab */
    int iterblock;
    MALLOC_INTERN(blocktab, SYMB_BLOKNBR*2, int);
    if (sopalin_data->sopar->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
      fprintf(stdout, "sizeof blocktab : %d integers\n", 2*SYMB_BLOKNBR);

    for (iterblock = 0; iterblock < SYMB_BLOKNBR; iterblock++)
      {
        blocktab[2*iterblock]   = SYMB_FROWNUM(iterblock);
        blocktab[2*iterblock+1] = SYMB_LROWNUM(iterblock);
      }
#  ifdef STARPU_BLOCKTAB_SELFCOPY
    {
      int ndevices;
      int device_id;
      CUDA_CALL(cudaGetDeviceCount(&ndevices));
      MALLOC_INTERN(d_blocktab, ndevices, int*);
      for (device_id = 0; device_id < ndevices; device_id++)
        {
          cudaSetDevice(device_id);
          CUDA_CALL(cudaMalloc((void*)&(d_blocktab[device_id]),
                               2*SYMB_BLOKNBR*sizeof(int)));
          CUDA_CALL(cudaMemcpy((void*)d_blocktab[device_id], blocktab,
                               2*SYMB_BLOKNBR*sizeof(int),
                               cudaMemcpyHostToDevice));
        }
    }
#  else
    starpu_vector_data_register(&blocktab_handle, 0,
                                (uintptr_t)blocktab, 2*SYMB_BLOKNBR,
                                sizeof(int));
    starpu_data_set_sequential_consistency_flag(blocktab_handle, 0);
#  endif
  }

#  ifdef PASTIX_DUMP_FACTO
  dump_all(datacode,
           ((datacode->updovct.sm2xtab!=NULL)?
            (DUMP_CSC | DUMP_SOLV | DUMP_SMB):(DUMP_CSC | DUMP_SOLV)));
#  endif

  thread_data = sopalin_data->thread_data[0];
  sopalin_data->sopar->diagchange = 0;
  SOPALIN_CLOCK_INIT;
#  ifdef STARPU_USE_CUDA
  /* starpu_helper_cublas_init(); */
#  endif


  nbpivots = 0;
  starpu_variable_data_register(&pivot_cnt_handle, 0,
                                (uintptr_t)&(nbpivots), sizeof (INT));
  starpu_data_set_reduction_methods(pivot_cnt_handle,
                                    &redux_codelet, &init_codelet);

#  ifdef CHOL_SOPALIN
  starpu_vector_data_register(&WORK_handle, -1, (uintptr_t)NULL, SOLV_COEFMAX,
                              sizeof(FLOAT));
#  else
  starpu_vector_data_register(&WORK_handle, -1, (uintptr_t)NULL, 2*SOLV_COEFMAX,
                              sizeof(FLOAT));
#  endif

  MALLOC_INTERN(trf_args,    SYMB_CBLKNBR, starpu_trf_data_t);

  MALLOC_INTERN(L_handle,    SYMB_CBLKNBR, starpu_data_handle_t);

#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
  MALLOC_INTERN(U_handle,    SYMB_CBLKNBR, starpu_data_handle_t);
#    endif
#  endif

  MALLOC_INTERN(gemm_args,      SYMB_BLOKNBR, starpu_gemm_data_t);


  {
    int itercblk;
    int max_cblksize = 0;
    int max_cblkcolnbr = 0;
    for (itercblk = 0; itercblk < SYMB_CBLKNBR; itercblk++)
      {
	max_cblksize   = MAX(max_cblksize,
			   CBLK_COLNBR(itercblk)*SOLV_STRIDE(itercblk));
	max_cblkcolnbr = MAX(max_cblkcolnbr,
			     CBLK_COLNBR(itercblk));
      }

    if (sopalin_data->sopar->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
      fprintf(stdout, "Maximum cblk size %d, maximu cblk colnbr %d\n",
              max_cblksize, max_cblkcolnbr);
  }

#  ifdef USE_TASK_DEP
  MALLOC_INTERN(starpu_tasktab, 2*SYMB_BLOKNBR, struct starpu_task *);
#  endif
#  if (defined CHOL_SOPALIN && defined SOPALIN_LU)
  starpu_partition_data(sopalin_data, L_handle, U_handle);
#  else
  starpu_partition_data(sopalin_data, L_handle, NULL);
#  endif

  SOPALIN_CLOCK_STOP;
  if (sopalin_data->sopar->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
    fprintf(stdout,"----- Time after data registration %lf s\n",
            SOPALIN_CLOCK_GET);

  /* For all column blocks we add a diag+trsm task.
     For all bloc in column block, we add a gemm task.
  */
  for (itertask=0;itertask<SOLV_TASKNBR;itertask++)
    {
      INT itercblk = TASK_CBLKNUM(itertask);
      struct starpu_task *task_diag = starpu_task_create();
      INT ret;
      INT iterbloc;
      INT handle_idx;
      /* We compute diagonal factorisation and TRSM */
      trf_args[itercblk].cblknum      = itercblk;
      trf_args[itercblk].sopalin_data = sopalin_data;

      task_diag->cl = &cl_trf;
      task_diag->cl_arg = &(trf_args[itercblk]);
      task_diag->destroy = 0;
      handle_idx = 0;

      task_diag->handles[handle_idx++] = pivot_cnt_handle;
      task_diag->handles[handle_idx++] = L_handle[itercblk];
#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
      /* LU */
      task_diag->handles[handle_idx++] = U_handle[itercblk];
#    endif
#  else /* CHOL_SOPALIN */
      /* LDLT */
      task_diag->handles[handle_idx++] = WORK_handle;
#  endif

      ASSERT(handle_idx == cl_trf.nbuffers, MOD_SOPALIN);

#  if (defined USE_TASK_DEP)
      {
        INT gcblk2list = UPDOWN_GCBLK2LIST(UPDOWN_LOC2GLOB( itercblk ));
        INT browk      = ( gcblk2list != -1 )?
          UPDOWN_LISTPTR( gcblk2list    ):-1;
        INT browk1     = ( gcblk2list != -1 )?
          UPDOWN_LISTPTR( gcblk2list + 1):-1;
        INT ndeps, iter;
        ndeps = browk1-browk;

        starpu_tasktab[SYMB_BLOKNUM(itercblk)] = task_diag;
        if (ndeps > deps_tab_size)
          {
            memFree_null(deps);
            deps_tab_size = ndeps;
            MALLOC_INTERN(deps, ndeps, struct starpu_task *);
          }
        for (iter = 0; iter < ndeps; iter++)
          {
            deps[iter] = starpu_tasktab[UPDOWN_LISTBLOK(browk+iter)];
          }
        if (ndeps != 0)
          starpu_task_declare_deps_array(task_diag, ndeps, deps);
      }
#  endif /* USE_TASK_DEP */

      ret = starpu_task_submit(task_diag);
      STARPU_ASSERT(!ret);

      for (iterbloc = SYMB_BLOKNUM(itercblk)+1;
           iterbloc < SYMB_BLOKNUM(itercblk+1);
           iterbloc ++)
        {
          struct starpu_task * task_gemm = starpu_task_create();
          INT blocnbr;
          blocnbr = SYMB_BLOKNUM(itercblk+1) - iterbloc;
          /* We compute GEMM */
          gemm_args[iterbloc].cblknum      = itercblk;
          gemm_args[iterbloc].bloknum      = iterbloc;
          gemm_args[iterbloc].nblocs       = blocnbr;
          gemm_args[iterbloc].fcblknum     = SYMB_CBLKNUM(iterbloc);
          gemm_args[iterbloc].sopalin_data = sopalin_data;
#  if (defined STARPU_BLOCKTAB_SELFCOPY)
	  gemm_args[iterbloc].d_blocktab   = d_blocktab;
#  endif

          task_gemm->cl = &cl_gemm;
          task_gemm->cl_arg = &(gemm_args[iterbloc]);
          task_gemm->destroy = 0;
          handle_idx = 0;

          task_gemm->handles[handle_idx++] = L_handle[itercblk];
          task_gemm->handles[handle_idx++] = L_handle[SYMB_CBLKNUM(iterbloc)];
#  if (defined CHOL_SOPALIN)
#    ifdef SOPALIN_LU
          task_gemm->handles[handle_idx++] = U_handle[itercblk];
          task_gemm->handles[handle_idx++] = U_handle[SYMB_CBLKNUM(iterbloc)];
#    endif
#  endif /* CHOL_SOPALIN */
          task_gemm->handles[handle_idx++] = WORK_handle;
#  ifndef STARPU_BLOCKTAB_SELFCOPY
          task_gemm->handles[handle_idx++] = blocktab_handle;
#  endif /* STARPU_BLOCKTAB_SELFCOPY */

          ASSERT(handle_idx == cl_gemm.nbuffers, MOD_SOPALIN);

#  ifdef USE_TASK_DEP
          starpu_tasktab[iterbloc] = task_gemm;
          deps[0] = task_diag;
          starpu_task_declare_deps_array(task_gemm, 1, deps);
#  endif /* USE_TASK_DEP */
          ret = starpu_task_submit(task_gemm);
          STARPU_ASSERT(!ret);
        }
    }

  /* Wait for all tasks to be finished */
  starpu_task_wait_for_all();
  SOPALIN_CLOCK_STOP;
  if (sopalin_data->sopar->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
    fprintf(stdout,"----- submission and wait for all %lf s\n",
            SOPALIN_CLOCK_GET);

#  if ( defined USE_TASK_DEP)
  memFree_null(deps);
#  endif

  /* Unregister buffers and leave starpu */
#  if(defined STARPU_PROFILING && defined USE_TASK_DEP)
  if (sopalin_data->sopar->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
    {
      int worker;
      /* Display the occupancy of all workers during the test */

      for (worker = 0; worker < starpu_worker_get_count(); worker++)
        {
          struct starpu_worker_profiling_info worker_info;
          int ret = starpu_worker_get_profiling_info(worker, &worker_info);
          STARPU_ASSERT(!ret);

          double total_time     = starpu_timing_timespec_to_us(&worker_info.total_time);
          double executing_time = starpu_timing_timespec_to_us(&worker_info.executing_time);
          double sleeping_time  = starpu_timing_timespec_to_us(&worker_info.sleeping_time);

          float executing_ratio = 100.0*executing_time/total_time;
          float sleeping_ratio  = 100.0*sleeping_time/total_time;

          char workername[128];

          double delay_sum[2]  = {0.0, 0.0};
          double length_sum[2] = {0.0, 0.0};
          unsigned int cnt[2]  = {0,   0};

          for (itertask=0;itertask<SOLV_TASKNBR;itertask++)
            {
              INT itercblk = TASK_CBLKNUM(itertask);
              INT iterbloc;
              struct starpu_task_profiling_info *info;
              info = starpu_tasktab[SYMB_BLOKNUM(itercblk)]->profiling_info;
              if (info->workerid == worker)
                {
                  /* How much time did it take before the task started ? */
                  delay_sum[0] += starpu_timing_timespec_delay_us(&info->submit_time,
                                                                  &info->start_time);

                  /* How long was the task execution ? */
                  length_sum[0] += starpu_timing_timespec_delay_us(&info->start_time,
                                                                   &info->end_time);
                  cnt[0]++;
                }

              for (iterbloc = SYMB_BLOKNUM(itercblk)+1;
                   iterbloc < SYMB_BLOKNUM(itercblk+1);
                   iterbloc ++)
                {
                  info = starpu_tasktab[iterbloc]->profiling_info;
                  if (info->workerid == worker)
                    {
                      /* How much time did it take before the task started ? */
                      delay_sum[1] += starpu_timing_timespec_delay_us(&info->submit_time,
                                                                      &info->start_time);

                      /* How long was the task execution ? */
                      length_sum[1] += starpu_timing_timespec_delay_us(&info->start_time,
                                                                       &info->end_time);
                      cnt[1]++;

                    }
                }
            }
          starpu_worker_get_name(worker, workername, 128);
          fprintf(stdout, "Worker %s:\n", workername);
          if (cnt[0] != 0)
            {
              fprintf(stdout, "Avg. delay on XXTRF : %2.2lf us, %d tasks\n",
                      (delay_sum[0])/cnt[0], cnt[0]);
              fprintf(stdout, "Avg. length on XXTRF : %2.2lf us\n",
                      (length_sum[0])/cnt[0]);
            }
          if (cnt[1] != 0)
            {
              fprintf(stdout, "Avg. delay on XXMM : %2.2lf us, %d tasks\n",
                      (delay_sum[1])/cnt[1], cnt[1]);
              fprintf(stdout, "Avg. length on XXMM : %2.2lf us\n",
                      (length_sum[1])/cnt[1]);
            }


          fprintf(stdout, "\ttotal time : %.2lf ms\n", total_time*1e-3);
          fprintf(stdout, "\texec time  : %.2lf ms (%.2f %%)\n", executing_time*1e-3, executing_ratio);
          fprintf(stdout, "\tblocked time  : %.2lf ms (%.2f %%)\n", sleeping_time*1e-3, sleeping_ratio);
        }
    }
#  endif /* USE_TASK_DEP && STARPU_PROFILING*/

  for (itertask=0;itertask<SOLV_TASKNBR;itertask++)
    {
      INT itercblk = TASK_CBLKNUM(itertask);
      INT iterbloc;
      for (iterbloc = SYMB_BLOKNUM(itercblk);
           iterbloc < SYMB_BLOKNUM(itercblk+1);
           iterbloc ++)
        {
          starpu_task_destroy(starpu_tasktab[iterbloc]);
        }
      starpu_data_unregister(L_handle[itercblk]);
#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
      starpu_data_unregister(U_handle[itercblk]);
#    endif
#  endif
    }
  starpu_data_unregister(WORK_handle);
  memFree_null(starpu_tasktab);

#  ifdef STARPU_BLOCKTAB_SELFCOPY
  {
    int ndevices;
    int device_id;
    CUDA_CALL(cudaGetDeviceCount(&ndevices));
    for (device_id = 0; device_id < ndevices; device_id++)
      {
	cudaSetDevice(device_id);
	CUDA_CALL(cudaFree(d_blocktab[device_id]));
      }
    memFree_null(d_blocktab);
  }
#  else
  starpu_data_unregister(blocktab_handle);
#  endif /* STARPU_BLOCKTAB_SELFCOPY */
  memFree_null(blocktab);

  starpu_data_unregister(pivot_cnt_handle);
  sopalin_data->sopar->diagchange = nbpivots;
#  ifdef STARPU_USE_CUDA
  /*starpu_helper_cublas_shutdown();*/
#  endif

  SOPALIN_CLOCK_STOP;
  if (sopalin_data->sopar->iparm[IPARM_VERBOSE] > API_VERBOSE_NO)
    fprintf(stdout,"----- sopalin time %lf\n",
            SOPALIN_CLOCK_GET);
  sopalin_data->sopar->dparm[DPARM_FACT_TIME] = SOPALIN_CLOCK_GET;
#  ifdef PASTIX_DUMP_FACTO
  dump_all(datacode,
           ((datacode->updovct.sm2xtab!=NULL)?
            (DUMP_CSC | DUMP_SOLV | DUMP_SMB):(DUMP_CSC | DUMP_SOLV)));
#  endif
  memFree_null(trf_args);
  memFree_null(gemm_args);
  memFree_null(L_handle);
#  ifdef CHOL_SOPALIN
#    ifdef SOPALIN_LU
  memFree_null(U_handle);
#    endif /* CHOL_SOPALIN */
#  endif /* SOPALIN_LU   */
  return NO_ERR;
}
#else
/* ISO C forbids an empty source file */
#  include "not_empty.h"
NOT_EMPTY(starpu_submit_tasks)
#endif
